#!/usr/bin/env python3
"""
=============================================================================
Make a Mac OS X App bundle (folder), using py2app.
Based on the more-complicated build.py of PyEdit.
Main file: run me in this folder to build the app.

An app allows text files to be associated to auto-open in PyEdit on clicks;
makes the program immune from Python changes; requires no Python installs;
and better supports file drag-and-drops, program icons, and more (but also
does a woefully-good job of hiding user-config source files and auto-saves!).
PyInstaller builds macos apps too, but py2app was chosen here early on.

Mac app also required... a mac-format .icns icon (the latter became an 
iconify.py extension), and fixfrozenpaths.py's context configurations.

Besides the executable, the app needs:
-MailConfigs: its files must be source code, and user-visible and editable;
-icon for the exe/app (window borders/app bar for Windows/Linux);
-UserGuide.html and docetc items it uses at runtime;
 
py2app sets the cwd to the bundle's Contents/Resource folder in all cases:
add '.' to sys.path, and use files copied to .=Resources, to which __file__
will refer.  Associations are irrelevant and unused here.

NOTE: it's assumed that the frozen app, plus the Python and Tk included 
in the app bundle are univeral binaries, supporting both 64- and 32-bit
machines.  This appears to be so, though the app's portability to older 
OS X versions remains to be shown (10.11 El Capitan is build host).

NOV21 UPDATE: fix paths (/MY-STUFF, /Users/blue) => (~/MY-STUFF, /Users/me)
for builds on new macOS Catalina machine, clarified a "mergeall" ref in the
docs, change one "/python3.5/" => "/python3.8/".  All else worked as is on 
Catalina with py2app 0.26.  Apps were rebuilt to pick up a newer Python's 
(3.8 on macos) support for newer openSSL; else some email servers now refuse
requests that have worked with the older app (and its Python 3.5) since 2017.

JUN2022 UPDATE: drop nested TextEditor/_private_ from shrinkpix project.
=============================================================================
"""

import os, sys, shutil
join, sep = os.path.join, os.path.sep
force = len(sys.argv) > 1                 # remake icon if any arg
startdir = os.getcwd()                    # this build script's dir

# 'python3' fails in both IDLE and PyEdit RunCode (env's PATH not inherited?)
python = '/usr/local/bin/python3'  # sys.executable (else proxy in PyEdit?)

#----------------------------------------------------------------------------
# make app's icon if one doesn't already exist
#----------------------------------------------------------------------------

iconship = join('..', '..', '..', 'icons')
iconmake = join('..', '..', 'build-icons')
iconname = 'pymailgui'
iconfile = iconname + '.icns'

# step into icon build dir and make
if force or not os.path.exists(iconship + sep + iconfile):
    os.chdir(iconmake)
   #os.system('iconutil -c icns %s.iconset' % iconname)                    # Apple util
   #os.system('./resize-on-mac.sh Pyedit1024new images-pyedit')            # just once
    os.system('%s iconify.py -mac images-pyedit %s' % (python, iconname))  # iconify 2.0
    os.chdir(startdir)
    shutil.move(join(iconmake, iconfile), join(iconship, iconfile))     # mv to ship

#----------------------------------------------------------------------------
# first, copy source tree to temp folder to avoid accidental code loss;
# [update - setup and teardown steps are now automated (run this script in 
# its dir), and Info.plist edits are now automatic by setup.py options;]
#----------------------------------------------------------------------------

# automated setup - run in this file's dir
temp = '/Users/me/Desktop/tempsrc'                      # cp can't include self!
if os.path.exists(temp):
    shutil.rmtree(temp)
os.mkdir(temp)

# move all to temp (no PP4E+TextEditor moves needed)
shutil.copytree('../../../../pymailgui', temp+'/pymailgui', symlinks=True) 
shutil.copy('setup.py', temp+'/pymailgui')              # add setup script to root
os.chdir(temp+'/pymailgui')                             # goto temp build dir

#----------------------------------------------------------------------------
# Drop big and irrelevant folders, plus any personal email stuff
#----------------------------------------------------------------------------

# 1) strip the 400M+ nested PyEdit build dirs
pyeditbuild = 'PyMailGui-PP4E/PP4E/Gui/TextEditor/build'
shutil.rmtree(pyeditbuild)
os.mkdir(pyeditbuild)

dummy = open(join(pyeditbuild, 'README.txt'), 'w')
dummy.write('Content removed: see PyEdit for its build scripts.\n')
dummy.close()  # else in-use on rmtree non Win (Mac?)

# 2) pyedit auto-saves; pass on reinstating zipped examples in docetc/examples
autosave_pmg = 'PyMailGui-PP4E/__pyedit-autosaves__'
autosave_pe  = 'PyMailGui-PP4E/PP4E/Gui/TextEditor/__pyedit-autosaves__'

for asave in (autosave_pmg, autosave_pe):
    for item in os.listdir(asave):
        if item != 'README-autosaves.txt':
            itempath = join(asave, item)
            print('Removing', itempath)
            os.remove(itempath)

# 3) TempParts; pass on reinstating zipped example in docetc/examples
tempparts = 'PyMailGui-PP4E/TempParts'
for item in os.listdir(tempparts):
    if item not in ['README-TempParts.txt', 'DSC00565.JPG', 'DSC03890.JPG']:
        itempath = join(tempparts, item)
        print('Removing', itempath)
        os.remove(itempath)       

# 4) sent mail; very dangerous...
print('Replacing sentmail')
sentmail = 'PyMailGui-PP4E/sentmail.txt'
os.remove(sentmail)
os.rename('PyMailGui-PP4E/sentmail-save.txt', sentmail)   # i.e., move 

# 5) Etc - nested PyEdit's non-public folder (just junk, really)
shutil.rmtree('PyMailGui-PP4E/PP4E/Gui/TextEditor/__private__')
shutil.rmtree('PyMailGui-PP4E/PP4E/Gui/TextEditor/_private_')     # (June2022: shrinkpix)

# 6) zip up large SentMail folder for space (it's 25M unzipped and 18M zipped,
# and will be difficult to select and open in the app version of PyMailGUI: Tk's
# file open dialog does not descend into app folders - a bit of a Tk buglet?)
#
code = '~/MY-STUFF/Code/ziptools/link'
os.chdir('PyMailGui-PP4E')  # for zip paths
thezip = thedir = 'SavedMail'
os.system('%s %s/zip-create.py %s %s' % (python, code, thezip, thedir)) 
shutil.rmtree(thedir)
os.chdir('..')   # back to temp/pymailgui dir

#----------------------------------------------------------------------------
# cp so use name for both the app-bundle folder and auto-menu title;
# this name must also be used in this project's setup.py file here;
# we can't rename these after the build, and the launcher is ingrained;
#----------------------------------------------------------------------------

# shutil.copy('textEditor.py', 'PyEdit.py')
# not here: rename after launcher app built ahead

#----------------------------------------------------------------------------
# build app-bundle folder in ./dist, using ./setup.py, copying N extras
# to the main Resources folder along with the generated program itself;
# this uses --resources to copy non-code extras, and -extras-scripts to
# also freeze related scripts shipped with [mergeall], so that they can
# be run from the app's Content/MacOS folder without a ".py" extension 
# and without requiring a separate Python install.
#
# Update: --extras-scripts is UNUSED here, as there are no other scripts.
# Instead, we use --package to bundle everything needed for the main
# PyMailGui program into the app's Python executable used to run the
# launcher.  This way, the launcher can spawn PyMailGui as source code
# (included in Content/Resources manually) using the app's bundled Python.
# That is, all PyMailGui.py requirements are "baked in" to the app's Python
# used by the frozen launcher executable.  Running PyMailGUI as source also
# allows its subfolder structure to remain intact, which is important for
# the MailConfigs top-level folder scheme.
#
# Most other source is for its docs only -- it may not run as source.
# Exception: the embedded PyEdit code can be run standalone in this app
# with a locally-installed Python, because it's included in full source
# form (but get the PyEdit app for a better user experience).
#----------------------------------------------------------------------------

extras = [                          # tbd: use a 'tools' folder?
    'MailConfigs',                  # ship these in Content/Resources='.'
    'README.txt', 
    'icons',
    'docetc',
    'UserGuide.html',
    'PyMailGui-PP4E',               # stripped above
    'Launch_PyEdit.pyw'             # runs as source if full PyMailGUI pkg included
]

alsofreeze = [                      # bake-in other scripts' requirements
    #'PyMailGui-PP4E/PyMailGui.py'  # not required: full --package bundled/scanned
]

exitstat = os.system(
    '%s setup.py py2app' 
    '   --iconfile icons/pymailgui.icns'
    '   --resources %s'
    '   --package PyMailGui-PP4E'
    % (python, ','.join(extras))
)

if exitstat:
    print('ERROR: build failed:', exitstat)
    sys.exit(exitstat)   # don't continue here

#----------------------------------------------------------------------------
# New: delete the redundant source-code copy of package PyMailGui-PP4E 
# added by py2app in the app's generated folder Resources/lib/python3.5.
# (Nov21: this is still the same, but named python3.8.)
#  
# This full copy is not needed, because we manually copy and run in a 
# PyMailGui-PP4E copy in Resources/ added via --resource.  The extra
# /lib copy takes up 63.8M in the unzipped app, and all its stdlib 
# requirements have been "baked in" to the bundle's app via --packages 
# (which really means that all its imported stdlib modules wind up in
# Resources/lib/python35.zip, which is on the app's sys.path).  We need 
# to use a live copy in Resources because parts need to be user-visible.
# Some launcher imports that were running before fixfrozenpath's '.'
# path extension had to be reordered to make this scheme work.
#
# The alternatives are to strip down the live copy in Resources; trim 
# parts of the /lib copy that are not essential (see below); avoid
# the Resource copy and run out of /lib; or accept an app that's 200M
# unzipped 147M zipped.  After the deletion here the app's size was
# reduced to 123M unzipped and 89M zipped -- still not small, but not
# much compared to Microsoft Office apps that top out at 2G installed.
# (Note that all sizes listed here are subject to change by release.)
#----------------------------------------------------------------------------

# we're in the temp/pymailgui folder root
print('Removing redundant PyMailGui-PP4E Resource/lib copy')
libroot = 'dist/Launch_PyMailGUI.app/Contents/Resources/lib/python3.8/PyMailGui-PP4E'
shutil.rmtree(libroot)

"""
The following alternative reduced size to 134M zipped and 97M unzipped;
dropping the rudundant package copy complete trimmed an extra ~10M.

print('Removing redundant data items in PyMailGui-PP4E /lib copy')
nukees = ('media',                                # 3M, but * 2 if kept
          'SavedMail.zip',                        # already zipped above, but 18M
          'PP4E/Gui/TextEditor/docetc/docimgs'    # 31M nested PyEdit's help images
         )

# drop nonessentials in package copy folder
for nukee in nukees:
    nukepath = join(libroot, nukee)
    print('Pruning', nukepath)
    if os.path.isdir(nukepath):
        shutil.rmtree(nukepath)
    else:
        os.remove(nukepath)
"""

#----------------------------------------------------------------------------
# cleanup: move and zip the app folder for easy xfer and backup (it has
# _very_ many files: nearly 3K files+folders for the current PyEdit App);
# [update - teardown actions are now automated (but still no data to copy)]
# don't copy extras to Contents/Resources folder here: automatic via the 
# py2app args above;  fixfrozenpaths.py arranges to see these as needed;
# DON'T -skipcruft in the zip command: py2app makes a Resources/site.pyc!
#----------------------------------------------------------------------------

# zip command: use portable ziptools (vs: 'zip -r %s %s' % (thezip, thedir))
thedir = 'PyMailGUI.app'
thezip = thedir + '.zip'
code   = '~/MY-STUFF/Code/ziptools/link'
zipit  = '%s %s/zip-create.py ../%s %s' % (python, code, thezip, thedir)

# move dist product folder here
os.chdir(startdir)
if os.path.exists('dist'):
    shutil.rmtree('dist')                      # nuke bogus retained temp?
shutil.move(temp+'/pymailgui/dist', '.')
shutil.rmtree(temp)                            # rm temp build tree 

# zip the nested app folder - unzip to test and use here or elsewhere
if os.path.exists(thezip):
    shutil.move(thezip, 'prev-'+thezip)        # save previous version
if os.path.exists(thedir):
    shutil.rmtree(thedir)                      # nuke unzipped version

os.chdir('dist')   
shutil.move('Launch_PyMailGUI.app', thedir)

os.system(zipit)                               # run zip in dist: has app dir
os.chdir('..')   
shutil.rmtree('dist')                          # rm dist: _very_ many files

print('Done: see ./%s' % thezip)

# +unzip app and copy it to /Applications to make it official

